/* ********************************************************************
 * Copyright (c) 2021 Biological Modeling and Visualization Laboratory,
 * University of Calgary. All rights reserved.
 * ********************************************************************/

/* Descriptive model of Canola (Brassica napus L.) */
#include <array>
#include <algorithm>
#include <cmath>
#include <lpfgall.h>

/* Declaration of constants */

#define DT 1. // simulation time step
#define STEPS int(val(MaxPlantAge) / DT);

#define DX 0.05 // step size for the leaf geometry
#define DX2 0.1 // step size for the petal geometry

/* Mapping of functions from functions.fset and functions.tset to macros */

#define branch_angle(n)     func(BRANCH_ANGLE,n/(val(MaxPhytomers)-val(VegPhytomers)))
#define branch_tropism(n)   func(BRANCH_TROPISM,n/val(VegPhytomers))

#define int_length(t)       tfunc(INT_LENGTH,t)
#define int_target_len(n)   func(INT_TARGET_LEN,n/val(MaxPhytomers))

#define leaf_length(t)      tfunc(LEAF_LENGTH,t)
#define leaf_target_len(n)  func(LEAF_TARGET_LEN,n/val(VegPhytomers))
#define leaf_target_wid(n)  func(LEAF_WIDTH_POS,n/val(VegPhytomers)) // NOTE: this function was set to 1 in the paper
#define leaf_inclination(t) tfunc(LEAF_INCLINATION,t) // NOTE: this function was set to 1 in the paper
#define leaf_width(x)       func(LEAF_WIDTH,x)
#define leaf_bend(x)        (360.0*func(LEAF_BEND,x))
#define leaf_angle(n)       func(LEAF_ANGLE,n/val(VegPhytomers))

#define pedicel_length(t)   tfunc(PEDICEL_LENGTH,t)

#define petal_angle(t)      tfunc(PETAL_ANGLE,t)
#define petal_length(t)	    tfunc(PETAL_LENGTH,t)
#define petal_width(x)      func(PETAL_WIDTH,x)

#define pod_length(t)       tfunc(POD_LENGTH,t)
#define prob_podding(t)     tfunc(PROB_PODDING,t/val(MaxPlantAge))

#define bud_vigour(n)       func(BUD_VIGOUR,n/val(VegPhytomers))
#define bud_delay(n)        func(BUD_DELAY,n/val(VegPhytomers))
#define bud_init_node(n)    func(BUD_INIT_NODE,n/val(VegPhytomers))

#define plastochron(n)      func(PLASTOCHRON,n/val(VegPhytomers))

/* Structure declarations and initialization */

typedef std::array<float,10> std_array; // typedef so it can be passed via lpfg's modules
std_array initArray = {0,0,0,0,0,0,0,0,0,0};

void UpdateAges (float sim_time, float age, std_array &ages) {
    if (sim_time <= 11.) // 2019-02-04 == day 11
        ages[0] = age;
    else if (sim_time <= 21.) 
        ages[1] = age;
    else if (sim_time <= 22.)
        ages[2] = age;
    else if (sim_time <= 24.)
        ages[3] = age;
    else if (sim_time <= 25.)
        ages[4] = age;
    else if (sim_time <= 27.)
        ages[5] = age;
    else if (sim_time <= 29.)
        ages[6] = age;
    else if (sim_time <= 33.)
        ages[7] = age;
    else if (sim_time <= 35.)
        ages[8] = age;
    else if (sim_time <= 39.)
        ages[9] = age;
}

struct OrganData {
	float age;
	float n; // NOTE: changed to 'float' from 'int' to avoid type casting 
	int order;	
    int state;

    std_array ages; // store organ age at different simulation times

	OrganData()
	  : age(0), n(0), order(0), state(0), ages(initArray)
	{}
};

OrganData apex_init_state;

/* Module definitions */

module A(OrganData); // apex
module I(OrganData); // internode
module L(OrganData); // leaf
module P(OrganData); // petiole
module K(OrganData); // flower
module C(OrganData); // petal
module D(OrganData); // seed pod
module B(OrganData); // axillary bud
module M(OrganData); // metamer for producing the above modules
module VisualAid(float,std_array);

/* Global variables */

float sim_time = 0.;

/* Organization of computation into phases */

Start: {
    sim_time = 0.;
    UseView(view001);
    UseView(view002);
    UseView(view003);
    UseView(view004);
    UseView(view005);
    UseView(view006);
    UseView(view007);
    UseView(view008);
    UseView(view009);
    UseView(view010);
}

EndEach: {
    sim_time += DT;
}

derivation length: STEPS;

Axiom: SB RollL(30) Right(-10.0)
          SetWidth(0.05) SetColor(1)
          SetElasticity(0,0.15)
          A(apex_init_state)
       EB
       VisualAid(0.,initArray);

production:

A(od) : {
	od.age += DT;
    UpdateAges(sim_time,od.age,od.ages);
	produce A(od);
}
B(od) : {
	od.age += DT;
    UpdateAges(sim_time,od.age,od.ages);
	produce B(od);
}
I(od) : {
	od.age += DT;
    UpdateAges(sim_time,od.age,od.ages);
	produce I(od);
}
L(od) : {
	od.age += DT;
    UpdateAges(sim_time,od.age,od.ages);
	produce L(od);
}
P(od) : {
	od.age += DT;
    UpdateAges(sim_time,od.age,od.ages);
	produce P(od);
}
K(od) : {
	od.age += DT;
    UpdateAges(sim_time,od.age,od.ages);
	produce K(od);
}
D(od) : {
	od.age += DT;
    UpdateAges(sim_time,od.age,od.ages);
	produce D(od);
}

VisualAid(age,ages) : {
    age += DT;
    UpdateAges(sim_time,age,ages);
    produce VisualAid(age,ages);
}

decomposition:
maximum depth: 200;

// define apical states
#define vegetative 0
#define reproductive 1
#define inactive 2

// define the apex behaviour
A(od) : {
	// model change of apex state
	if (od.state == vegetative && od.n >= val(VegPhytomers))
		od.state = reproductive;
	else if (od.state == reproductive && od.n >= val(MaxPhytomers))
		od.state = inactive;
    // check if apex should produce a phytomer
	if (od.state != inactive) { 
        // calculate plastochron given apex state
        float p;
        if (od.state == vegetative)
            // NOTE: the third term, (1.-val(ShortV)*od.order), was set to one in the paper,
            // but is here to give extra control over the vegetative plastochron
            p = val(VegPlastochron) * plastochron(od.n) * (1.-val(ShortV)*od.order);
        else
            p = val(RepPlastochron);
        // if plastochron has elapsed
        if (od.age >= p) {
            od.age = od.age - p;
            nproduce M(od) // produce a phytomer
                     RollL(137.5); // simulate phyllotaxis
            od.n += 1; // then increase the phytomer count
            produce A(od);
        }
	}
}

M(od) : {
    od.ages = initArray;
    nproduce I(od); // an internode
	if (od.state == vegetative) {
        produce SB B(od) EB  // an axillary bud, and
                SB L(od) EB; // a leaf
	}
	else { // in reproductive state
		produce SB Down(55.) P(od) K(od) EB; // NOTE: In the paper, the pedicel angle, Down(55), was missing
	}
}

B(od) : {
    // ignore buds above maximum branch order
    if (od.order >= val(MaxOrder) || val(SHOW_BRANCHES) == 0) 
        produce;
    // calculate bud vigour
    float vigour = bud_vigour(od.n) * (1. - val(ShortV) * od.order);
	if (vigour > val(MinBudVigour)) {
        // set the branching angle and the shoot gravitropic
        // response before new lateral apex is produced
        nproduce Down(val(BranchAngle) * branch_angle(od.n))
                 SetElasticity(0,val(Elast) * branch_tropism(od.n));
        // set organ data to produce new lateral apex
		od.state = vegetative;
		od.age = val(BudDelay) * bud_delay(od.n);
		od.order = od.order + 1;
		od.n = std::floor(val(VegPhytomers) * bud_init_node(od.n));
        od.ages = initArray;
		produce A(od);
	}
    // NOTE: produce B(od); was incorrectly specified in the paper
}

/* // Moved to interpretation for time calibration model
K(od): {
     if (od.age > PoddingAge)
          if (ran(1) < prob_podding(t/MaxPlantAge)) {
               od.age = 0;    // set pod age to zero
               produce D(od); // flower becomes a pod
          } else produce;     // or flower aborts
     produce K(od);           // otherwise flower remains
}
*/

interpretation:
maximum depth: 200;

#define InternodeInterpretation(agesIndex) \
I(od) : { \
    float shorten = 1. - val(ShortInt) * od.order; \
    float len = val(IntLen) * int_target_len(od.n) * shorten * int_length(od.ages[agesIndex]); \
    if (len <= 0.f) \
        produce; \
    float rad = std::pow(len,val(ExpIntRad)); \
    int color = od.n >= val(VegPhytomers) ? 4 : 3; \
	produce SetColor(color) SetWidth(rad) F(len); \
}

#define LeafInterpretation(agesIndex) \
L(od) : { \
    if (val(SHOW_LEAVES)==0)\
        produce;\
	float len = val(LeafLen) * leaf_target_len(od.n) \
        * (1. - val(ShortLeaf) * od.order) * leaf_length(od.ages[agesIndex]); \
    if (len <= 0.f) \
        produce; \
	float wid = val(LeafWid) * leaf_target_wid(od.n) * std::pow(len,val(ExpLeafWid)); \
	nproduce SetElasticity(0,0) SetColor(7) \
            SetWidth(wid * leaf_width(0)) \
		    Down(val(BranchAngle) * branch_angle(od.n) * leaf_inclination(od.ages[agesIndex])) \
		    CurrentContour(1) \
            StartGC(); \
    float x; \
    for (x = 0.; x < 1.-DX; x += DX) { \
        nproduce F(len*DX) \
                 SetWidth(wid * leaf_width(x)) \
                 Down(leaf_bend(x) * leaf_angle(od.n) * DX); \
    } \
    nproduce SetWidth(wid * leaf_width(1.)) F(len*(1.-x)); \
    produce EndGC(); \
}

#define PedicleInterpretation(agesIndex) \
P(od) : { \
    if (val(SHOW_FLOWERS)==0)\
        produce;\
	float len = val(PedicelLen) * std::pow(int_target_len(od.n),val(ExpPedicelLen)) * pedicel_length(od.ages[agesIndex]); \
    float rad = std::pow(len,val(ExpPedicelRad)); \
	produce SetWidth(rad) F(len); \
}

#define FlowerInterpretation(agesIndex) \
K(od) : { \
    if (val(SHOW_FLOWERS)==0)\
        produce;\
\
    sran(od.n); \
	if (od.ages[agesIndex] > val(PoddingAge)) { \
        if (ran(1) < prob_podding(sim_time)) { \
		    od.age = 0.0;\
		    produce SetElasticity(0,0.05) D(od);\
        }\
	}\
\
    nproduce SetElasticity(0,0) SetColor(7); \
    for (int i = 0; i < 4; i++) { \
        nproduce RollL(90. * float(i)) SB C(od) EB; \
    } \
    produce; \
} \
C(od): { \
    float len = val(PetalLen) * std::pow(int_target_len(od.n),val(ExpPetalLen)) * petal_length(od.ages[agesIndex]); \
    float wid = std::pow(len,val(ExpPetalWid)); \
    nproduce Down(180. * petal_angle(od.ages[agesIndex])); \
    nproduce SetWidth(wid * petal_width(0)); \
    nproduce CurrentContour(2) SetColor(5); \
    nproduce StartGC(); \
    float x; \
    for (x = 0.; x < 1.-DX2; x += DX2) { \
        nproduce F(len*DX2) SetWidth(wid * petal_width(x)); \
    } \
    nproduce SetWidth(wid * petal_width(1.)) F(len*(1.-x)); \
    produce EndGC(); \
}

#define SiliqueInterpretation(agesIndex) \
D(od) : { \
    if (val(SHOW_FLOWERS)==0)\
        produce;\
    float len = val(PodLen) * std::pow(int_target_len(od.n),val(ExpPodLen)) * pod_length(od.ages[agesIndex]); \
    float rad = std::pow(len,val(ExpPodRad)); \
	produce SetColor(5) SB \
		   SetWidth(rad) StartGC F(len*0.8) \
		   SetWidth(0.1*rad) F(len*0.2) EndGC \
		   EB; \
}

// Draw image in the background
#define ViualAidInterpretation(day,view) \
VisualAid(age,ages) : { \
    static int NUM_OTHER_TEXS = 1; \
    const float SF = 3.; \
	float w = 614.0 * SF; \
	float h = 1028.0 * SF; \
    int tex = day; \
    nproduce SetElasticity(0,0.); \
    nproduce SetColor(1); \
    nproduce SetWidth(w); \
    nproduce CurrentContour(3) TextureVCoeff(1.0/h); \
    nproduce CurrentTexture(tex+NUM_OTHER_TEXS); \
    nproduce SB MoveTo(0.,-170.*SF,800.) StartGC() F(h) EndGC() EB; \
    produce; \
}

vgroup 0:

vgroup view001:

InternodeInterpretation(0)
LeafInterpretation(0)
PedicleInterpretation(0)
FlowerInterpretation(0)
SiliqueInterpretation(0)

ViualAidInterpretation(0,0)

vgroup view002:

InternodeInterpretation(1)
LeafInterpretation(1)
PedicleInterpretation(1)
FlowerInterpretation(1)
SiliqueInterpretation(1)

ViualAidInterpretation(1,1)

vgroup view003:

InternodeInterpretation(2)
LeafInterpretation(2)
PedicleInterpretation(2)
FlowerInterpretation(2)
SiliqueInterpretation(2)

ViualAidInterpretation(2,2) 

vgroup view004:

InternodeInterpretation(3)
LeafInterpretation(3)
PedicleInterpretation(3)
FlowerInterpretation(3)
SiliqueInterpretation(3)

ViualAidInterpretation(3,3) 

vgroup view005:

InternodeInterpretation(4)
LeafInterpretation(4)
PedicleInterpretation(4)
FlowerInterpretation(4)
SiliqueInterpretation(4)

ViualAidInterpretation(4,3) 

vgroup view006:

InternodeInterpretation(5)
LeafInterpretation(5)
PedicleInterpretation(5)
FlowerInterpretation(5)
SiliqueInterpretation(5)

ViualAidInterpretation(5,3) 


vgroup view007:

InternodeInterpretation(6)
LeafInterpretation(6)
PedicleInterpretation(6)
FlowerInterpretation(6)
SiliqueInterpretation(6)

ViualAidInterpretation(6,3) 


vgroup view008:

InternodeInterpretation(7)
LeafInterpretation(7)
PedicleInterpretation(7)
FlowerInterpretation(7)
SiliqueInterpretation(7)

ViualAidInterpretation(7,3) 


vgroup view009:

InternodeInterpretation(8)
LeafInterpretation(8)
PedicleInterpretation(8)
FlowerInterpretation(8)
SiliqueInterpretation(8)

ViualAidInterpretation(8,3) 


vgroup view010:

InternodeInterpretation(9)
LeafInterpretation(9)
PedicleInterpretation(9)
FlowerInterpretation(9)
SiliqueInterpretation(9)

ViualAidInterpretation(9,3) 
